import RockPaperScissorsFullGame.RPS as RPS
import RockPaperScissorsFullGame.ML as ML
# import RPS
# import ML

def main():
    # arr = [{1 : {'player': 'Paper', 'computer': 'Rock'}}, {2 : {'player': 'Scissors', 'computer': 'Paper'}}, {3 : {'player': 'Rock', 'computer': 'Scissors'}}, {4 : {'player': 'Paper', 'computer': 'Paper'}}]
    # for kvp in arr :
    #     ML.rpsStorage.update(kvp)
    ML.printRps()
    #print welcome banner
    print("""    
      __        __   __         __   ___  __     __   __   __          __        __   ___  __     __   __     __   __   __   __   __  
|  | |__  |    /  ` /  \  |\/| |__    |  /  \   |__) /  \ /  ` |__/   |__)  /\  |__) |__  |__)   /__` /  ` | /__` /__` /  \ |__) /__` 
|/\| |__  |___ \__, \__/  |  | |__    |  \__/   |  \ \__/ \__, |  \   |    /--\ |    |___ |  \   .__/ \__, | .__/ .__/ \__/ |  \ .__/                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           
    """)
    
    #first ask user if they want to see ascii art
    asciiVisibility = RPS.setAsciiVisibility()
    
    #Prompting game difficulty
    difficulty = ML.chooseDifficulty()
    if (difficulty.value == 1) :
        print("You chose easy difficulty! Computer will be on random mode.") 
    elif (difficulty.value == 2) :
        print("You chose hard difficulty! Computer will be using history matching.")

    # if user enters difficulty greater than easy, ask user if they want to turn on text-detail mode (to see how computer tracks history)
    # by default, detail mode is set to false
    # showDetails will be passed to the ML.py functions that are responsible for printing out the logic behind the computer move
    showDetails = False
    if (difficulty.value > 1):
        details = input("Do you want to turn on detail mode?\nselect one\n(1) Yes\n(2) No\n")
        if (details == "1" or details.lower() == "yes"): 
            showDetails = True
        elif (details == "2" or details.lower() == "no"):
            showDetails = False
        else:
            print("Invalid input... defaulting detailed mode to false")

    # game
    play = True
    round = 1
    while (play):
        #Initialize computer choice (computerItem) to None during the start of every round.
        computerItem = None

        #Easy mode: randomizes and returns an rps item for the computer choice.
        if (difficulty.value == 1) :
            computerItem = RPS.getComputerItem()

        #Medium mode: calls ML_historyMatching() to make the computer choice.
        elif (difficulty.value == 2) :
            #Current round has to be at least 3 since minimum past rounds checked is 2.
            if round >= 3 :
                #If there is a successful prediction, the computer choice will be chosen to counter the predicted player item.
                # pass showDetails... if true then the func will print out 
                predictedPlayerItem = (ML.ML_historyMatching("computer", showDetails))
                if (predictedPlayerItem  != None) :
                    #only print predicted move if in text-detail mode
                    if(showDetails):
                        print("predicted move: ", predictedPlayerItem) #
                    if (predictedPlayerItem['player'] == 'Scissors') :
                        computerItem = RPS.Item.Rock
                    elif (predictedPlayerItem['player'] == 'Paper') :
                        computerItem = RPS.Item.Scissors
                    elif (predictedPlayerItem['player'] == 'Rock') :
                        computerItem = RPS.Item.Paper
            
            #If there are no successful predictions, the computer choice will default to random.
            if (computerItem == None) :
                computerItem = RPS.getComputerItem()

        #Prompts player input.
        playerInput = input("Enter your item: (rock, paper, scissors) ")

        #Allows the player to end the program with "quit" as the input.
        if(playerInput.lower() == "quit"):
            print("Thank you for playing!\n")
            break

        else:
            #Assigns an rps item corresponding to the user input.
            playerItem = RPS.inputToItem(playerInput)

            #If user input doesn't correspond to any rps item, prompts user input
            #and restarts the program.
            if (playerItem == None):
                print("Invalid Input\n")
                continue

            else:
                playerMove = RPS.getPlayerAsciiArt(playerItem)
                computerMove = RPS.getComputerAsciiArt(computerItem)
                if(not asciiVisibility):
                    print("You choose " + playerItem.name + ". Computer chooses " + computerItem.name + ".")
                else:
                    print("Player Move: " + playerItem.name + playerMove)
                    print("Computer Move: " + computerItem.name + computerMove)
                result = RPS.getOutcome(playerItem,computerItem)
                print("You " + result.name + ".\n")

                #Offering history matching
                # print(ML.ML_callHistoryMatching("player"))

                #Print rpsStorage
                ML.printRps(showDetails)
        
        #Tracking the round
        round += 1


if __name__ == '__main__':
    main()


from enum import Enum
import random
# import ML
import RockPaperScissorsFullGame.ML as ML


#Denotes instances of Item (plays) with values assigned to them.
class Item(Enum):
    Rock = 1
    Paper = 2
    Scissors = 3

#Denotes instances of Outcome (results) with values assigned to them.
class Outcome(Enum):
    Win = 1
    Lose = 2
    Tie = 3

#Accepts two choices (player and computer) and outputs the outcome of the round.
def getOutcome(playerItem, computerItem):
    if(playerItem.value == computerItem.value):
        ML.storeOutcome("Tie", playerItem.name, computerItem.name)
        return Outcome.Tie
    elif((playerItem == Item.Rock and computerItem == Item.Scissors) or (playerItem.value == computerItem.value + 1)):
        ML.storeOutcome("Win", playerItem.name, computerItem.name)
        return Outcome.Win
    else:
        ML.storeOutcome("Lose", playerItem.name, computerItem.name)
        return Outcome.Lose

#Randomizes a number between 1 and 3, and returns the Item (rock, paper, scissors)
#associated with the random number.
def getComputerItem():
    index = random.randint(1,3)
    return Item(index)

#Checks if the input is in the beginning of rock, paper, or scissors.
#Returns appropriate instance associated with the input.
def inputToItem(str):
    if ("rock").startswith(str.lower()):
        return Item.Rock
    elif ("paper").startswith(str.lower()):
        return Item.Paper
    elif ("scissors").startswith(str.lower()):
        return Item.Scissors
    else:
        return None
    
    
def setAsciiVisibility():
    seeAscii = input("Do you want ascii art to be visible?\nselect one\n(1) Yes\n(2) No\n")
    if (seeAscii == "1" or seeAscii.lower() == "yes"): 
        return True
    elif (seeAscii == "2" or seeAscii.lower() == "no"):
        return False
    else:
        print("Invalid input... defaulting ascii visibility to true")
        return True   

def getPlayerAsciiArt(item):
    if(item == Item.Rock):
        return """
            _____
        ---'  ____)  
             (_____)  
             (_____)  
             (____)
        ---.(___)
        """
    elif (item == Item.Paper):
        return """
            ______
        ---'   ____)____  
                   ______)  
                   _______)  
                  _______)
        ---._________) 
        """
    elif (item == Item.Scissors):
        return """
            ______
        ---'   ____)_____ 
                    ______)  
                __________)  
              (____)
        ---.__(___)
        """
    else:
        return "Invalid Option"

def getComputerAsciiArt(item):
    if(item == Item.Rock):
        return """
           _____
         (____  '---
        (_____)  
        (_____)  
         (____)
           (___).---

        """
    elif (item == Item.Paper):
        return """
             _____
       ____(____  '---
     (_____ 
    (_______  
     (_______
       (_________.---
        """
    elif (item == Item.Scissors):
        return """
                _____
         ______(____ '--- 
       (______  
       (__________  
            (____)
             (___)__.---
        """
    else:
        return "Invalid Option"


#Machine Learning functions for higher difficulties
# import RPS
import RockPaperScissorsFullGame.RPS as RPS
from enum import Enum

#Stores the results of each round in a dictionary: 
#{round : {"player" : playerchoice, "computer" :computerchoice}}.
rpsStorage = {}

#Arrays track the rounds for which the outcome occurs.
winsArr = []
lossesArr = []
tiesArr = []

#Variables track the number of rounds and outcomes.
round = 1
wins = 0
losses = 0
ties = 0

#Difficulties
class difficulty(Enum):
    Easy = 1
    Hard = 2

#Prompt the user for choice of difficulty and returns difficulty 
#if the difficulty begins with the input.
def chooseDifficulty() :
    playerInput = input("Enter your preferred difficulty (easy, hard): ")
    if ("easy").startswith(playerInput.lower()): 
        return difficulty.Easy
    elif ("hard").startswith(playerInput.lower()): 
        return difficulty.Hard
    else:
        return None

#Stores the outcome of a round and players' choices (player and computer)
#in the rpsStorage dictionary. Tracks wins, losses, and ties.
def storeOutcome(outcome, playerChoice, computerChoice) :
    global round 
    global wins 
    global losses 
    global ties 

    if (outcome == "Win") :
        kvp = {round : storeChoices(playerChoice, computerChoice)}
        rpsStorage.update(kvp)
        winsArr.append(round)
        round += 1
        wins += 1
        return kvp
    elif (outcome == "Lose") :
        kvp = {round : storeChoices(playerChoice, computerChoice)}
        rpsStorage.update(kvp)
        lossesArr.append(round)
        round += 1
        losses += 1
        return kvp
    elif (outcome == "Tie") :
        kvp = {round : storeChoices(playerChoice, computerChoice)}
        rpsStorage.update(kvp)
        tiesArr.append(round)
        round += 1
        ties += 1
        return kvp
    else :
        return None

#Auxiliary function: Creates the choices dictionary for rpsStorage.
def storeChoices(playerChoice, computerChoice) :
    return {"player" : playerChoice, "computer" : computerChoice}

#Prints the current outcomes in rpsStorage 
# showDetails defaults to true so that it does not impact unit tests
def printRps(showDetails=True) :
    #If there aren't any values in RPS storage (shouldn't ever be the case), prints 0.
    # only print if in text-detail mode
    if(showDetails):
        if len(rpsStorage) == 0:
            print(0)
        else :
            tracker = 1

            print("\nOutcomes of the previous rounds: ")

            #Printing the round and the value in the rpsStorage for the outcome of each round.
            for kvp in rpsStorage.values() :
                print(tracker, " : ", kvp)
                tracker += 1

            print("")

#Iterates through rpsStorage to check if the outcomes from a previous number
#of rounds before the current one has been repeated before. Returns the next
#expected outcome after the repeated interval of outcomes if there is a repeated
#segment. If there isn't a repeated segment, the previous number of rounds checked
#is decremented and checked again. This repetition occurs until number of rounds
#is lower than 2 (not really a segment).
# showDetails defaults to true so that it does not impact unit tests
def ML_historyMatching (numRounds, showDetails=True) :
    #Fetching the values in rpsStorage dictionary.
    resultsArr = list(rpsStorage.values())

    while numRounds != 1 :
        fractResultsArr = []

        #Iterates through the last n rounds and stores their key-value pairs into fractResultsArr[]. 
        for index in range(len(resultsArr) - numRounds, len(resultsArr), 1) :
            fractResultsArr.append(resultsArr[index])

        #Using the auxiliary function find_subarray(), a segment of key-value pairs in the larger resultsArr[] 
        #containing the key-value pairs of the smaller fractResultsArr[] will be found. An index will be returned
        #that will serve as the starting index for the segment of key-value pairs in resultsArr[] that match the 
        #ones in fractResultsArr[].
        startIndex = find_subarray(resultsArr, fractResultsArr)
        
        # only print if in text-detail mode
        if(showDetails):
            print("this is the fract results arr: ", fractResultsArr) #
            print("this is the start index: ", startIndex) #

        #If there are no segments of key-value pairs in the larger resultsArr[] which match the ones in
        #the smaller fractResultsArr[], the number of past rounds checked will be decremented. This loop continues
        #until past rounds equals 1. 
        if startIndex == None :
            numRounds -= 1
            continue

        #Returns the expected move the player will play after the matched segment. Checks if the expected move is in
        #the bounds of the resultsArr (numRounds can't encompass the entire array AND the matching segment cannot be last n 
        #rounds).  
        elif not (startIndex + numRounds >= len(resultsArr)):
            return resultsArr[startIndex + numRounds]

        numRounds -= 1
    return None

#Auxiliary function to call ML_historyMatching and prompt the user for num rounds input.
def ML_callHistoryMatching(caller, showDetails=True) :
    global round

    #Prompts player input if the caller is the player
    if ("player").startswith(caller.lower()) :
        playerInput = input("Would you like to check the history (yes or no)?\n")

        #If the player's response is yes, prompts for how many past rounds to check.
        if ("yes").startswith(playerInput.lower()) :
            playerInput = int(input("How many rounds?\n"))

            #Checks if the player wants to check an appropriate number of past rounds (makes sure the current round is more than the number of past rounds).
            if (playerInput >= 2 and round > playerInput) :
                return ML_historyMatching(playerInput, showDetails)
            else :
                return ValueError
        
        #Additional output if user doesn't want to do history matching.
        elif ("no").startswith(playerInput.lower()) :
            print("Next time!")
    
    #If the caller is the computer, automatically sets the numRounds to 1 less than the total rounds in the game.
    elif ("computer").startswith(caller.lower()) :
        if (round >= 3) :
            return ML_historyMatching(round-2, showDetails)
    
    return None
      
#Checks if the first array contains the second array.
def find_subarray(first_arr, second_arr):
    first = 0
    second = 0

    first_len = len(first_arr)
    second_len = len(second_arr)

    while first < first_len and second < second_len:
        #From the beginning of each array, every element is matched to check
        #for an equality. In this case, we only want to match the player's past moves
        #to find a pattern so only the key-value pair for the player is checked (hence [0]).
        if (first_arr[first])['player'] == (second_arr[second])['player']:
            first += 1
            second += 1

            #If the second array is completed (has every element equal
            #to an element in the first array in the same order), 
            #second array is contained within the first array.
            if second == second_len:
                return first - second_len

        else:
            #If the second array's element doesn't match the first array's 
            #element, it has to be reset. Likewise, if the second array is
            #reset, the first array has to be checked again from the next 
            #element after the first element which matched the second array's 
            #first element.
            first = first - second + 1
            second = 0

    return None


        
    




    
    
    

